﻿// Author: Pietro Polsinelli - http://designAGame.eu
// Twitter https://twitter.com/ppolsinelli
// All free as in free beer :-)

using System;
using System.Collections.Generic;

namespace OL
{
    /// <summary>
    /// The rationale behind this class is presented here:
    /// https://designagame.eu/2017/05/on-jon-ingolds-how-unhappy-is-unhappy
    /// </summary>
    public class CharacterTrait<T>
    {
        readonly float oldEpisodesDecay = 1;

        readonly List<Boundary<T>> boundaries;

        readonly List<CharacterEpisode> allEpisodes = new List<CharacterEpisode>();

        /// <param name="stateBoundariesOnPositiveEpisodes">
        /// A list of % of happyness and resulting character trait state label
        /// </param>
        /// <param name="oldEpisodesDecay">
        /// The % of decay to be applied to all older episodes once a new one happens. 
        /// Should be something like 0.97
        /// </param>
        public CharacterTrait(Dictionary<float, T> stateBoundariesOnPositiveEpisodes,
            float oldEpisodesDecay = 1)
        {
            boundaries = new List<Boundary<T>>();
            foreach (KeyValuePair<float, T> pair in stateBoundariesOnPositiveEpisodes)
            {
                if (!pair.Key.Betweenf(0, 1))
                    throw new Exception("Must use % as boundaries");
                boundaries.Add(new Boundary<T> { BoundaryValue = pair.Key,Label = pair.Value });
            }
            boundaries.Sort((a, b) => b.BoundaryValue.CompareTo(a.BoundaryValue));

            this.oldEpisodesDecay = oldEpisodesDecay;            
        }

        public CharacterTrait<T> PositiveEpisode(float impact = 1, string log = null)
        {
            AddLog(true, log, impact);
            return this;
        }

        public CharacterTrait<T> NegativeEpisode(float impact = 1, string log = null)
        {
            AddLog(false, log, impact);
            return this;
        }

        void AddLog(bool positive, string log, float impact)
        {
            foreach (CharacterEpisode episode in new List<CharacterEpisode>(allEpisodes))
            {
                episode.DecayedImpact *= oldEpisodesDecay;
            }
            allEpisodes.Add(new CharacterEpisode()
            {
                Positive = positive,
                Log = log,
                Impact = impact,
                DecayedImpact = impact
            });
        }

        public float HowMuchPositivityInPerc()
        {
           float positiveEpisodesTotal = 0;
            float negativeEpisodesTotal = 0;
            foreach (CharacterEpisode episode in allEpisodes)
            {
                if (episode.Positive)
                {
                    positiveEpisodesTotal += episode.DecayedImpact;
                }
                else
                {
                    negativeEpisodesTotal += episode.DecayedImpact;
                }
            }
            return positiveEpisodesTotal / (positiveEpisodesTotal + negativeEpisodesTotal);
        }

        public float HowMuchNegativityInPerc()
        {
            float positiveEpisodesTotal = 0;
            float negativeEpisodesTotal = 0;
            foreach (CharacterEpisode episode in allEpisodes)
            {
                if (episode.Positive)
                {
                    positiveEpisodesTotal += episode.DecayedImpact;
                }
                else
                {
                    negativeEpisodesTotal += episode.DecayedImpact;
                }
            }
            return negativeEpisodesTotal / (positiveEpisodesTotal + negativeEpisodesTotal);            
        }

        public T State()
        {
            float pPerc = HowMuchPositivityInPerc();
            
            foreach (Boundary<T> boundary in boundaries)
            {
                if (pPerc >= boundary.BoundaryValue)
                {
                    return boundary.Label;
                }
            }
            return default(T);
        }

        public struct Boundary<T>
        {
            public float BoundaryValue;
            public T Label;
        }

        public class CharacterEpisode
        {
            public bool Positive;
            public string Log;
            public float Impact; // original value
            public float DecayedImpact;
        }
    }
}